@@ -9,7 +9,7 @@
 // where we later scan or directly feed data into API's.
 static int _process_tokenize(mcp_parser_t *pr, const size_t max) {
     const char *s = pr->request;
-    int len = pr->reqlen - 2;
+    int len = pr->endlen;
 
     // since multigets can be huge, we can't purely judge reqlen against this
     // limit, but we also can't index past it since the tokens are shorts.
@@ -93,7 +93,7 @@ static int _process_request_key(mcp_parser_t *pr) {
 // Returns the offset for the next key.
 size_t _process_request_next_key(mcp_parser_t *pr) {
     const char *cur = pr->request + pr->parsed;
-    int remain = pr->reqlen - pr->parsed - 2;
+    int remain = pr->endlen - pr->parsed;
 
     // chew off any leading whitespace.
     while (remain) {
@@ -126,7 +126,7 @@ static int _process_request_metaflags(mcp_parser_t *pr, int token) {
         return 0;
     }
     const char *cur = pr->request + pr->tokens[token];
-    const char *end = pr->request + pr->reqlen - 2;
+    const char *end = pr->request + pr->endlen;
 
     // We blindly convert flags into bits, since the range of possible
     // flags is deliberately < 64.
@@ -266,8 +266,12 @@ static int _process_request_storage(mcp_parser_t *pr, size_t max) {
 }
 
 // common request with key: <cmd> <key> <args>
-static int _process_request_simple(mcp_parser_t *pr, const size_t max) {
+static int _process_request_simple(mcp_parser_t *pr, const int min, const int max) {
     _process_tokenize(pr, max);
+    if (pr->ntokens < min) {
+        P_DEBUG("%s: not enough tokens for simple request: %d\n", __func__, pr->ntokens);
+        return -1;
+    }
     pr->keytoken = 1; // second token is usually the key... stupid GAT.
 
     _process_request_key(pr);
@@ -290,15 +294,25 @@ int process_request(mcp_parser_t *pr, const char *command, size_t cmdlen) {
         return -1;
     }
 
-    const char *s = memchr(command, ' ', cmdlen-2);
+    // Commands can end with bare '\n's. Depressingly I intended to be strict
+    // with a \r\n requirement but never did this and need backcompat.
+    // In this case we _know_ \n is at cmdlen because we can't enter this
+    // function otherwise.
+    if (cm[cmdlen-2] == '\r') {
+        pr->endlen = cmdlen - 2;
+    } else {
+        pr->endlen = cmdlen - 1;
+    }
+
+    const char *s = memchr(command, ' ', pr->endlen);
     if (s != NULL) {
         cl = s - command;
     } else {
-        cl = cmdlen - 2;
+        cl = pr->endlen;
     }
     pr->keytoken = 0;
     pr->has_space = false;
-    pr->parsed = cl + 1;
+    pr->parsed = cl;
     pr->request = command;
     pr->reqlen = cmdlen;
     int token_max = PARSER_MAX_TOKENS;
@@ -351,7 +365,7 @@ int process_request(mcp_parser_t *pr, const char *command, size_t cmdlen) {
                     cmd = CMD_GET;
                     type = CMD_TYPE_GET;
                     token_max = 2; // don't chew through multigets.
-                    ret = _process_request_simple(pr, 2);
+                    ret = _process_request_simple(pr, 2, 2);
                 }
                 if (cm[1] == 'a' && cm[2] == 't') {
                     type = CMD_TYPE_GET;
@@ -375,13 +389,13 @@ int process_request(mcp_parser_t *pr, const char *command, size_t cmdlen) {
                 cmd = CMD_GETS;
                 type = CMD_TYPE_GET;
                 token_max = 2; // don't chew through multigets.
-                ret = _process_request_simple(pr, 2);
+                ret = _process_request_simple(pr, 2, 2);
             } else if (strncmp(cm, "incr", 4) == 0) {
                 cmd = CMD_INCR;
-                ret = _process_request_simple(pr, 4);
+                ret = _process_request_simple(pr, 3, 4);
             } else if (strncmp(cm, "decr", 4) == 0) {
                 cmd = CMD_DECR;
-                ret = _process_request_simple(pr, 4);
+                ret = _process_request_simple(pr, 3, 4);
             } else if (strncmp(cm, "gats", 4) == 0) {
                 cmd = CMD_GATS;
                 type = CMD_TYPE_GET;
@@ -393,7 +407,7 @@ int process_request(mcp_parser_t *pr, const char *command, size_t cmdlen) {
         case 5:
             if (strncmp(cm, "touch", 5) == 0) {
                 cmd = CMD_TOUCH;
-                ret = _process_request_simple(pr, 4);
+                ret = _process_request_simple(pr, 3, 4);
             } else if (strncmp(cm, "stats", 5) == 0) {
                 cmd = CMD_STATS;
                 // Don't process a key; fetch via arguments.
@@ -406,7 +420,7 @@ int process_request(mcp_parser_t *pr, const char *command, size_t cmdlen) {
         case 6:
             if (strncmp(cm, "delete", 6) == 0) {
                 cmd = CMD_DELETE;
-                ret = _process_request_simple(pr, 4);
+                ret = _process_request_simple(pr, 2, 4);
             } else if (strncmp(cm, "append", 6) == 0) {
                 cmd = CMD_APPEND;
                 ret = _process_request_storage(pr, token_max);
@@ -440,7 +454,7 @@ int process_request(mcp_parser_t *pr, const char *command, size_t cmdlen) {
 // FIXME (v2): any reason to pass in command/cmdlen separately?
 mcp_request_t *mcp_new_request(lua_State *L, mcp_parser_t *pr, const char *command, size_t cmdlen) {
     // reserving an upvalue for key.
-    mcp_request_t *rq = lua_newuserdatauv(L, sizeof(mcp_request_t) + MCP_REQUEST_MAXLEN * 2 + KEY_MAX_LENGTH, 1);
+    mcp_request_t *rq = lua_newuserdatauv(L, sizeof(mcp_request_t) + MCP_REQUEST_MAXLEN + KEY_MAX_LENGTH, 1);
     // TODO (v2): memset only the non-data part? as the rest gets memcpy'd
     // over.
     memset(rq, 0, sizeof(mcp_request_t));
@@ -449,7 +463,6 @@ mcp_request_t *mcp_new_request(lua_State *L, mcp_parser_t *pr, const char *comma
     memcpy(rq->request, command, cmdlen);
     rq->pr.request = rq->request;
     rq->pr.reqlen = cmdlen;
-    gettimeofday(&rq->start, NULL);
 
     luaL_getmetatable(L, "mcp.request");
     lua_setmetatable(L, -2);
@@ -459,64 +472,129 @@ mcp_request_t *mcp_new_request(lua_State *L, mcp_parser_t *pr, const char *comma
     return rq;
 }
 
-// TODO (v2):
-// if modified, this will re-serialize every time it's accessed.
-// a simple opt could copy back over the original space
-// a "better" one could A/B the request ptr and clear the modified state
-// each time it gets serialized.
-void mcp_request_attach(lua_State *L, mcp_request_t *rq, io_pending_proxy_t *p) {
+// fill a preallocated request object.
+void mcp_set_request(mcp_parser_t *pr, mcp_request_t *rq, const char *command, size_t cmdlen) {
+    memset(rq, 0, sizeof(mcp_request_t));
+    memcpy(&rq->pr, pr, sizeof(*pr));
+
+    memcpy(rq->request, command, cmdlen);
+    rq->pr.request = rq->request;
+    rq->pr.reqlen = cmdlen;
+}
+
+// Replaces a token inside a request and re-parses.
+// Note that this has some optimization opportunities. Delaying until
+// required.
+// We should not guarantee order when updating meta flags, which would allow
+// blanking tokens and appending new ones.
+// TODO (v2): much of the length is the key, avoid copying it.
+int mcp_request_render(mcp_request_t *rq, int idx, const char flag, const char *tok, size_t len) {
+    char temp[MCP_REQUEST_MAXLEN+1];
+    char *p = temp;
+    mcp_parser_t *pr = &rq->pr;
+
+    if (pr->reqlen + len + 2 > MCP_REQUEST_MAXLEN) {
+        return -1;
+    }
+    // Cannot add/append tokens yet.
+    if (idx >= pr->ntokens) {
+        return -1;
+    }
+
+    memcpy(p, pr->request, pr->tokens[idx]);
+    p += pr->tokens[idx];
+
+    if (flag) {
+        *p = flag;
+        p++;
+    }
+    if (tok) {
+        memcpy(p, tok, len);
+        p += len;
+    }
+
+    // Add a space and copy more tokens if there were more.
+    if (idx+1 < pr->ntokens) {
+        if (flag || len != 0) {
+            // Only pre-space if not deleting the token.
+            *p = ' ';
+            p++;
+        }
+        memcpy(p, &pr->request[pr->tokens[idx+1]], pr->tokens[pr->ntokens] - pr->tokens[idx+1]);
+        p += pr->tokens[pr->ntokens] - pr->tokens[idx+1];
+    } else {
+        // If we removed something from the end we might've left some spaces.
+        while (*(p-1) == ' ') {
+            p--;
+        }
+    }
+
+    memcpy(p, "\r\n\0", 3);
+    p += 2;
+
+    memcpy(rq->request, temp, p - temp);
+
+    // Hold the vlen/vbuf and restore after re-parsing. Since we can only edit
+    // the command line, not the value here, we would otherwise allow sending
+    // arbitrary memory over the network if someone modifies a SET.
+    void *vbuf = pr->vbuf;
+    int vlen = pr->vlen;
+
+    memset(pr, 0, sizeof(mcp_parser_t)); // TODO: required?
+    int ret = process_request(pr, rq->request, p - temp);
+    if (ret != 0) {
+        return ret;
+    }
+    pr->vbuf = vbuf;
+    pr->vlen = vlen;
+    return 0;
+}
+
+int mcp_request_append(mcp_request_t *rq, const char flag, const char *tok, size_t len) {
+    mcp_parser_t *pr = &rq->pr;
+    const char *start = pr->request;
+    char *p = (char *)pr->request + pr->reqlen - 2; // start at the \r
+    assert(*p == '\r');
+
+    if (pr->reqlen + len + 2 > MCP_REQUEST_MAXLEN) {
+        return -1;
+    }
+
+    *p = ' ';
+    p++;
+
+    if (flag) {
+        *p = flag;
+        p++;
+    }
+    if (tok) {
+        memcpy(p, tok, len);
+        p += len;
+    }
+
+    memcpy(p, "\r\n\0", 3);
+    p += 2;
+
+    // See note on mcp_request_render()
+    void *vbuf = pr->vbuf;
+    int vlen = pr->vlen;
+
+    memset(pr, 0, sizeof(mcp_parser_t)); // TODO: required?
+    int ret = process_request(pr, rq->request, p - start);
+    if (ret != 0) {
+        return ret;
+    }
+    pr->vbuf = vbuf;
+    pr->vlen = vlen;
+
+    return 0;
+}
+
+void mcp_request_attach(mcp_request_t *rq, io_pending_proxy_t *p) {
     mcp_parser_t *pr = &rq->pr;
     char *r = (char *) pr->request;
     size_t len = pr->reqlen;
 
-    // one or more of the tokens were changed
-    if (rq->was_modified) {
-        assert(rq->tokent_ref);
-        // option table to top of stack.
-        lua_rawgeti(L, LUA_REGISTRYINDEX, rq->tokent_ref);
-
-        // space was reserved in case of modification.
-        char *nr = rq->request + MCP_REQUEST_MAXLEN;
-        r = nr;
-        char *or = NULL;
-
-        for (int x = 0; x < pr->ntokens; x++) {
-            const char *newtok = NULL;
-            size_t newlen = 0;
-            if (x != 0 && x != pr->keytoken) {
-                int type = lua_rawgeti(L, -1, x+1);
-                if (type != LUA_TNIL) {
-                    newtok = lua_tolstring(L, -1, &newlen);
-                    memcpy(nr, newtok, newlen);
-                    nr += newlen;
-                }
-                lua_pop(L, 1);
-            }
-
-            if (newtok == NULL) {
-                // TODO (v2): if we add an extra "end" token that's just reqlen we can
-                // memcpy... however most args are short and that may not be worth
-                // it.
-                or = rq->request + pr->tokens[x];
-                // will walk past the end without the \r test.
-                // if we add the end token trick this can be changed.
-                while (*or != ' ' && *or != '\r' && *or != '\n') {
-                    *nr = *or;
-                    nr++;
-                    or++;
-                }
-            }
-            *nr = ' ';
-            nr++;
-        }
-        // tag the end bits.
-        memcpy(nr-1, "\r\n\0", 3);
-        nr++;
-
-        len = nr - (rq->request + MCP_REQUEST_MAXLEN);
-        lua_pop(L, 1); // pop the table
-    }
-
     // The stringified request. This is also referencing into the coroutine
     // stack, which should be safe from gc.
     p->iov[0].iov_base = r;
@@ -529,12 +607,12 @@ void mcp_request_attach(lua_State *L, mcp_request_t *rq, io_pending_proxy_t *p)
         p->iovcnt = 2;
         p->iovbytes += pr->vlen;
     }
-
 }
 
 // second argument is optional, for building set requests.
 // TODO: append the \r\n for the VAL?
 int mcplib_request(lua_State *L) {
+    LIBEVENT_THREAD *t = PROXY_GET_THR(L);
     size_t len = 0;
     size_t vlen = 0;
     mcp_parser_t pr = {0};
@@ -583,8 +661,14 @@ int mcplib_request(lua_State *L) {
             proxy_lua_error(L, "failed to allocate value memory for request object");
         }
         memcpy(rq->pr.vbuf, val, vlen);
+        // Note: Not enforcing the memory limit here is deliberate:
+        // - if we're over the memory limit, it'll get caught very soon after
+        // this, but we won't be causing some lua to bail mid-flight, which is
+        // more graceful to the end user.
+        pthread_mutex_lock(&t->proxy_limit_lock);
+        t->proxy_buffer_memory_used += rq->pr.vlen;
+        pthread_mutex_unlock(&t->proxy_limit_lock);
     }
-    gettimeofday(&rq->start, NULL);
 
     // rq is now created, parsed, and on the stack.
     return 1;
@@ -651,49 +735,67 @@ int mcplib_request_token(lua_State *L) {
         return 1;
     }
 
-    // we hold overwritten or parsed tokens in a lua table.
-    if (rq->tokent_ref == 0) {
-        // create a presized table that can hold our tokens.
-        lua_createtable(L, rq->pr.ntokens, 0);
-        // duplicate value to set back
-        lua_pushvalue(L, -1);
-        rq->tokent_ref = luaL_ref(L, LUA_REGISTRYINDEX);
-    } else {
-        lua_rawgeti(L, LUA_REGISTRYINDEX, rq->tokent_ref);
-    }
-    // top of stack should be token table.
-
     size_t vlen = 0;
     if (argc > 2) {
         // overwriting a token.
-        luaL_checklstring(L, 3, &vlen);
-        lua_pushvalue(L, 3); // copy to top of stack
-        lua_rawseti(L, -2, token);
-        rq->was_modified = true;
+        size_t newlen = 0;
+        const char *newtok = lua_tolstring(L, 3, &newlen);
+        if (mcp_request_render(rq, token-1, 0, newtok, newlen) != 0) {
+            proxy_lua_error(L, "token(): request malformed after edit");
+            return 0;
+        }
         return 0;
     } else {
         // fetching a token.
-        if (lua_rawgeti(L, -1, token) != LUA_TSTRING) {
-            lua_pop(L, 1); // got a nil, drop it.
+        const char *start = rq->pr.request + rq->pr.tokens[token-1];
+        vlen = _process_token_len(&rq->pr, token-1);
 
-            // token not uploaded yet. find the len.
-            const char *start = rq->pr.request + rq->pr.tokens[token-1];
-            vlen = _process_token_len(&rq->pr, token-1);
-
-            P_DEBUG("%s: pushing token of len: %lu\n", __func__, vlen);
-            lua_pushlstring(L, start, vlen);
-            lua_pushvalue(L, -1); // copy
-
-            lua_rawseti(L, -3, token); // pops copy.
-        }
-
-        // return fetched token or copy of new token.
+        P_DEBUG("%s: pushing token of len: %lu\n", __func__, vlen);
+        lua_pushlstring(L, start, vlen);
         return 1;
     }
 
     return 0;
 }
 
+// Fetch only.
+int mcplib_request_token_int(lua_State *L) {
+    mcp_request_t *rq = luaL_checkudata(L, 1, "mcp.request");
+    int argc = lua_gettop(L);
+
+    if (argc == 1) {
+        lua_pushnil(L);
+        return 1;
+    }
+
+    int x = luaL_checkinteger(L, 2);
+
+    if (x < 1 || x > rq->pr.ntokens) {
+        // maybe an error?
+        lua_pushnil(L);
+        return 1;
+    }
+
+    size_t vlen = 0;
+    // fetching a token.
+    const char *s = rq->pr.request + rq->pr.tokens[x-1];
+    vlen = _process_token_len(&rq->pr, x-1);
+    // do a funny dance to safely strtol the token.
+    // TODO: use tokenizer based tokto when merged.
+    char temp[22];
+    int tocopy = vlen > 22 ? 21 : vlen;
+    memcpy(temp, s, tocopy);
+    temp[vlen] = '\0';
+    int64_t token = 0;
+    if (safe_strtoll(temp, &token)) {
+        lua_pushinteger(L, token);
+    } else {
+        lua_pushnil(L);
+    }
+
+    return 1;
+}
+
 int mcplib_request_ntokens(lua_State *L) {
     mcp_request_t *rq = luaL_checkudata(L, 1, "mcp.request");
     lua_pushinteger(L, rq->pr.ntokens);
@@ -745,12 +847,6 @@ int mcplib_request_flag_token(lua_State *L) {
     }
     if (lua_isstring(L, 3)) {
         // overwriting a flag/token with the third argument.
-        // NOTE: semi duplicated from mcplib_request_token()
-        if (rq->tokent_ref == 0) {
-            // create a presized table that can hold our tokens.
-            lua_createtable(L, rq->pr.ntokens, 0);
-            rq->tokent_ref = luaL_ref(L, LUA_REGISTRYINDEX);
-        }
         replace = true;
     }
     uint64_t flagbit = (uint64_t)1 << (flagstr[0] - 65);
@@ -772,14 +868,12 @@ int mcplib_request_flag_token(lua_State *L) {
 
                 // Have something to replace the flag/token with.
                 if (replace) {
-                    // table is top of stack.
-                    lua_rawgeti(L, LUA_REGISTRYINDEX, rq->tokent_ref);
-                    // need to copy the string to top of stack.
-                    lua_pushvalue(L, 3);
-
-                    lua_rawseti(L, -2, x+1); // pops copy
-                    lua_pop(L, 1); // pop table
-                    rq->was_modified = true;
+                    size_t newlen = 0;
+                    const char *newtok = lua_tolstring(L, 3, &newlen);
+                    if (mcp_request_render(rq, x, 0, newtok, newlen) != 0) {
+                        proxy_lua_error(L, "token(): request malformed after edit");
+                        return 0;
+                    }
                 }
                 break;
             }
@@ -791,19 +885,369 @@ int mcplib_request_flag_token(lua_State *L) {
     return ret;
 }
 
-int mcplib_request_gc(lua_State *L) {
-    mcp_request_t *rq = luaL_checkudata(L, -1, "mcp.request");
+// returns bool, int
+// bool results if flag exists or not
+// if int conversion fails, int is nil
+int mcplib_request_flag_token_int(lua_State *L) {
+    mcp_request_t *rq = luaL_checkudata(L, 1, "mcp.request");
+    size_t len = 0;
+    const char *flagstr = luaL_checklstring(L, 2, &len);
+    if (len != 1) {
+        proxy_lua_error(L, "has_flag(): meta flag must be a single character");
+        return 0;
+    }
+    if (flagstr[0] < 65 || flagstr[0] > 122) {
+        proxy_lua_error(L, "has_flag(): invalid flag, must be A-Z,a-z");
+        return 0;
+    }
+
+    uint64_t flagbit = (uint64_t)1 << (flagstr[0] - 65);
+
+    int ret = 1;
+    if (rq->pr.t.meta.flags & flagbit) {
+        lua_pushboolean(L, 1);
+        for (int x = rq->pr.keytoken+1; x < rq->pr.ntokens; x++) {
+            const char *s = rq->pr.request + rq->pr.tokens[x];
+            if (s[0] == flagstr[0]) {
+                size_t vlen = _process_token_len(&rq->pr, x);
+                if (vlen > 1) {
+                    // do a funny dance to safely strtol the token.
+                    // TODO: use tokenizer based tokto when merged.
+                    char temp[22];
+                    int tocopy = vlen > 22 ? 21 : vlen-1;
+                    memcpy(temp, s+1, tocopy);
+                    temp[vlen-1] = '\0';
+                    int64_t token = 0;
+                    if (safe_strtoll(temp, &token)) {
+                        lua_pushinteger(L, token);
+                    } else {
+                        lua_pushnil(L);
+                    }
+                    ret = 2;
+                }
+
+                break;
+            }
+        }
+    } else {
+        lua_pushboolean(L, 0);
+    }
+
+    return ret;
+}
+
+// these functions take token as string or number
+// if number, internally convert it to avoid creating garbage
+static inline char _mcp_request_get_arg_flag(lua_State *L, int idx) {
+    size_t len = 0;
+    const char *flagstr = luaL_checklstring(L, idx, &len);
+
+    if (len != 1) {
+        proxy_lua_error(L, "request: meta flag must be a single character");
+        return 0;
+    }
+    if (flagstr[0] < 65 || flagstr[0] > 122) {
+        proxy_lua_error(L, "request: invalid flag, must be A-Z,a-z");
+        return 0;
+    }
+
+    return flagstr[0];
+}
+
+// *tostring must be large enough to hold a 64bit number as a string.
+static inline const char * _mcp_request_check_flag_token(lua_State *L, int idx, char *tostring, size_t *tlen) {
+    const char *token = NULL;
+    *tlen = 0;
+    if (lua_isstring(L, idx)) {
+        token = lua_tolstring(L, idx, tlen);
+    } else if (lua_isnumber(L, idx)) {
+        int isnum = 0;
+        lua_Integer n = lua_tointegerx(L, idx, &isnum);
+        if (isnum) {
+            char *end = itoa_64(n, tostring);
+            token = tostring;
+            *tlen = end - tostring;
+        } else {
+            proxy_lua_error(L, "request: invalid flag argument");
+            return NULL;
+        }
+    } else if (lua_isnoneornil(L, idx)) {
+        // no token, just add the flag.
+    } else {
+        proxy_lua_error(L, "request: invalid flag argument");
+        return NULL;
+    }
+
+    return token;
+}
+
+// req:flag_add("F", token) -> (bool)
+// if token is "example", appends "Fexample" to request
+int mcplib_request_flag_add(lua_State *L) {
+    mcp_request_t *rq = luaL_checkudata(L, 1, "mcp.request");
+    char flag = _mcp_request_get_arg_flag(L, 2);
+    char tostring[30];
+
+    uint64_t flagbit = (uint64_t)1 << (flag - 65);
+    if (rq->pr.t.meta.flags & flagbit) {
+        // fail, flag already exists.
+        lua_pushboolean(L, 0);
+        return 1;
+    }
+
+    size_t tlen = 0;
+    const char *token = _mcp_request_check_flag_token(L, 3, tostring, &tlen);
+
+    if (mcp_request_append(rq, flag, token, tlen) == 0) {
+        lua_pushboolean(L, 1);
+    } else {
+        lua_pushboolean(L, 0);
+    }
+
+    return 1;
+}
+
+// req:flag_set("F", token) -> (bool) [overwrites if exists]
+// if token is "example", appends "Fexample" to request
+int mcplib_request_flag_set(lua_State *L) {
+    mcp_request_t *rq = luaL_checkudata(L, 1, "mcp.request");
+    char flag = _mcp_request_get_arg_flag(L, 2);
+    char tostring[30];
+
+    int x = mcp_request_find_flag_index(rq, flag);
+    size_t tlen = 0;
+    const char *token = _mcp_request_check_flag_token(L, 3, tostring, &tlen);
+
+    if (x > 0) {
+        // TODO: do nothing if:
+        // flag exists in request, without token, and we're not setting a
+        // token.
+        if (mcp_request_render(rq, x, flag, token, tlen) != 0) {
+            lua_pushboolean(L, 0);
+            return 1;
+        }
+    } else {
+        if (mcp_request_append(rq, flag, token, tlen) != 0) {
+            lua_pushboolean(L, 0);
+            return 1;
+        }
+    }
+
+    lua_pushboolean(L, 1);
+    return 1;
+}
+
+// allows replacing a flag with a different flag
+// req:flag_replace("F", "N", token) -> (bool)
+// if token is "example", appends "Nexample" to request
+int mcplib_request_flag_replace(lua_State *L) {
+    mcp_request_t *rq = luaL_checkudata(L, 1, "mcp.request");
+    char flag = _mcp_request_get_arg_flag(L, 2);
+    char newflag = _mcp_request_get_arg_flag(L, 3);
+    char tostring[30];
+
+    int x = mcp_request_find_flag_index(rq, flag);
+    size_t tlen = 0;
+    const char *token = _mcp_request_check_flag_token(L, 4, tostring, &tlen);
+
+    if (x > 0) {
+        if (mcp_request_render(rq, x, newflag, token, tlen) != 0) {
+            lua_pushboolean(L, 0);
+            return 1;
+        }
+    } else {
+        if (mcp_request_append(rq, newflag, token, tlen) != 0) {
+            lua_pushboolean(L, 0);
+            return 1;
+        }
+    }
+
+    lua_pushboolean(L, 1);
+    return 1;
+}
+
+// req:flag_del("F") -> (bool)
+// remove a flag if exists
+int mcplib_request_flag_del(lua_State *L) {
+    mcp_request_t *rq = luaL_checkudata(L, 1, "mcp.request");
+    char flag = _mcp_request_get_arg_flag(L, 2);
+
+    int x = mcp_request_find_flag_index(rq, flag);
+
+    if (x > 0) {
+        if (mcp_request_render(rq, x, 0, NULL, 0) != 0) {
+            lua_pushboolean(L, 0);
+            return 1;
+        }
+    } else {
+        // nothing there, didn't delete anything.
+        lua_pushboolean(L, 0);
+        return 1;
+    }
+
+    lua_pushboolean(L, 1);
+    return 1;
+}
+
+// local match, token = req:match_res(res)
+// checks if req has `k` or `O`. If so, checks response for `K` or `O`
+// returns true, nil if matches
+// returns false, res token if not match.
+//
+int mcplib_request_match_res(lua_State *L) {
+    mcp_request_t *rq = luaL_checkudata(L, 1, "mcp.request");
+    mcp_resp_t *rs = luaL_checkudata(L, 2, "mcp.response");
+
+    const char *opaque_token = NULL;
+    size_t opaque_len = 0;
+    mcmc_resp_t reresp;
+
+    // requests all have keys. check for an opaque.
+    mcp_request_find_flag_token(rq, 'O', &opaque_token, &opaque_len);
+    mcmc_parse_buf(rs->buf, rs->blen, &reresp);
+
+    // scan the response line for tokens, since we don't have a reciprocal API
+    // yet. When we do this code will be replaced with a function call like
+    // the above.
+    const char *p = reresp.rline;
+    const char *e = p + reresp.rlen;
+    if (!p) {
+        // happens if the result line is blank (ie; 'HD\r\n')
+        lua_pushboolean(L, 0);
+        lua_pushnil(L);
+        return 2;
+    }
+
+    int matched = 0;
+    while (p != e) {
+        if (*p == ' ') {
+            p++;
+        } else if (*p == 'k' || *p == 'O') {
+            const char *rq_token = NULL;
+            int rq_len = 0;
+            if (*p == 'k') {
+                rq_token = MCP_PARSER_KEY(rq->pr);
+                rq_len = rq->pr.klen;
+            } else if (*p == 'O') {
+                rq_token = opaque_token;
+                rq_len = opaque_len;
+            }
+            if (rq_token == NULL) {
+                lua_pushboolean(L, 0);
+                lua_pushnil(L);
+                return 2;
+            }
+
+            p++; // skip flag and start comparing token
+            const char *rs_token = p;
+
+            // find end of token
+            while (p != e && !isspace(*p)) {
+                p++;
+            }
+
+            int rs_len = p - rs_token;
+            if (rq_len != rs_len || memcmp(rq_token, rs_token, rs_len) != 0) {
+                // FAIL, keys aren't the same length or don't match.
+                lua_pushboolean(L, 0);
+                lua_pushlstring(L, rs_token, rs_len);
+                return 2;
+            } else {
+                matched = 1;
+            }
+        } else {
+            // skip token
+            while (p != e && *p != ' ') {
+                p++;
+            }
+        }
+    }
+
+    lua_pushboolean(L, matched);
+    lua_pushnil(L);
+    return 2;
+}
+
+void mcp_request_cleanup(LIBEVENT_THREAD *t, mcp_request_t *rq) {
     // During nread c->item is the malloc'ed buffer. not yet put into
     // rq->buf - this gets freed because we've also set c->item_malloced if
     // the connection closes before finishing nread.
     if (rq->pr.vbuf != NULL) {
+        pthread_mutex_lock(&t->proxy_limit_lock);
+        t->proxy_buffer_memory_used -= rq->pr.vlen;
+        pthread_mutex_unlock(&t->proxy_limit_lock);
         free(rq->pr.vbuf);
+        // need to ensure we NULL this out now, since we can call the cleanup
+        // routine independent of GC, and a later GC would double-free.
+        rq->pr.vbuf = NULL;
+    }
+}
+
+int mcplib_request_gc(lua_State *L) {
+    LIBEVENT_THREAD *t = PROXY_GET_THR(L);
+    mcp_request_t *rq = luaL_checkudata(L, -1, "mcp.request");
+    mcp_request_cleanup(t, rq);
+
+    return 0;
+}
+
+static int _mcp_request_find_flag(mcp_request_t *rq, const char flag) {
+    uint64_t flagbit = (uint64_t)1 << (flag - 65);
+    if (rq->pr.t.meta.flags & flagbit) {
+        for (int x = rq->pr.keytoken+1; x < rq->pr.ntokens; x++) {
+            const char *s = rq->pr.request + rq->pr.tokens[x];
+            if (s[0] == flag) {
+                return x;
+            }
+        }
+    }
+    return -1;
+}
+
+int mcp_request_find_flag_index(mcp_request_t *rq, const char flag) {
+    int x = _mcp_request_find_flag(rq, flag);
+    return x;
+}
+
+int mcp_request_find_flag_token(mcp_request_t *rq, const char flag, const char **token, size_t *len) {
+    int x = _mcp_request_find_flag(rq, flag);
+    if (x > 0) {
+        size_t tlen = _process_token_len(&rq->pr, x);
+        if (tlen > 1) {
+            *token = rq->pr.request + rq->pr.tokens[x] +1;
+        } else {
+            *token = NULL;
+        }
+        *len = tlen-1;
+    }
+    return x;
+}
+
+// FIXME: temporary copypasta accessor until request objects can be moved to
+// mcmc tokenizer.
+int mcp_request_find_flag_tokenint64(mcp_request_t *rq, const char flag, int64_t *token) {
+    for (int x = rq->pr.keytoken+1; x < rq->pr.ntokens; x++) {
+        const char *s = rq->pr.request + rq->pr.tokens[x];
+        if (s[0] == flag) {
+            size_t vlen = _process_token_len(&rq->pr, x);
+            if (vlen > 1) {
+                // do a funny dance to safely strtol the token.
+                char temp[22];
+                int tocopy = vlen > 22 ? 21 : vlen-1;
+                memcpy(temp, s+1, tocopy);
+                temp[vlen-1] = '\0';
+                if (safe_strtoll(temp, token)) {
+                    return 0;
+                } else {
+                    return -1;
+                }
+            }
+
+            break;
+        }
     }
 
-    if (rq->tokent_ref != 0) {
-        luaL_unref(L, LUA_REGISTRYINDEX, rq->tokent_ref);
-    }
-    return 0;
+    return -1;
 }
 
 // TODO (v2): check what lua does when it calls a function with a string argument
