@@ -8,6 +8,7 @@
  * memcached protocol.
  */
 #include "memcached.h"
+#include "storage.h"
 #include <sys/mman.h>
 #include <sys/stat.h>
 #include <sys/socket.h>
@@ -279,7 +280,10 @@ void slabs_init(const size_t limit, const double factor, const bool prealloc, co
     {
         char *t_initial_malloc = getenv("T_MEMD_INITIAL_MALLOC");
         if (t_initial_malloc) {
-            mem_malloced = (size_t)atol(t_initial_malloc);
+            int64_t env_malloced;
+            if (safe_strtoll((const char *)t_initial_malloc, &env_malloced)) {
+                mem_malloced = (size_t)env_malloced;
+            }
         }
 
     }
@@ -299,6 +303,10 @@ void slabs_prefill_global(void) {
     while (mem_malloced < mem_limit
             && (ptr = memory_allocate(len)) != NULL) {
         grow_slab_list(0);
+        // Ensure the front header is zero'd to avoid confusing restart code.
+        // It's probably good enough to cast it and just zero slabs_clsid, but
+        // this is extra paranoid.
+        memset(ptr, 0, sizeof(item));
         p->slab_list[p->slabs++] = ptr;
     }
     mem_limit_reached = true;
@@ -381,9 +389,11 @@ static int do_slabs_newslab(const unsigned int id) {
         return 0;
     }
 
-#if !defined(__FreeBSD__)
+    // Always wipe the memory at this stage: in restart mode the mmap memory
+    // could be unused, yet still full of data. Better for usability if we're
+    // wiping memory as it's being pulled out of the global pool instead of
+    // blocking startup all at once.
     memset(ptr, 0, (size_t)len);
-#endif
     split_slab_page_into_freelist(ptr, id);
 
     p->slab_list[p->slabs++] = ptr;
@@ -447,6 +457,8 @@ static void do_slabs_free_chunked(item *it, const size_t size) {
     it->prev = 0;
     // header object's original classid is stored in chunk.
     p = &slabclass[chunk->orig_clsid];
+    // original class id needs to be set on free memory.
+    it->slabs_clsid = chunk->orig_clsid;
     if (chunk->next) {
         chunk = chunk->next;
         chunk->prev = 0;
@@ -626,7 +638,7 @@ static void *memory_allocate(size_t size) {
 }
 
 /* Must only be used if all pages are item_size_max */
-static void memory_release() {
+static void memory_release(void) {
     void *p = NULL;
     if (mem_base != NULL)
         return;
@@ -992,7 +1004,7 @@ static int slab_rebalance_move(void) {
                         /* These are definitely required. else fails assert */
                         new_it->it_flags &= ~ITEM_LINKED;
                         new_it->refcount = 0;
-                        do_item_replace(it, new_it, hv);
+                        do_item_replace(it, new_it, hv, ITEM_get_cas(it));
                         /* Need to walk the chunks and repoint head  */
                         if (new_it->it_flags & ITEM_CHUNKED) {
                             item_chunk *fch = (item_chunk *) ITEM_schunk(new_it);
@@ -1314,6 +1326,7 @@ int start_slab_maintenance_thread(void) {
         fprintf(stderr, "Can't create rebal thread: %s\n", strerror(ret));
         return -1;
     }
+    thread_setname(rebalance_tid, "mc-slabmaint");
     return 0;
 }
 
